/*
 * /////////////////////////////////////////////////////////////////////////////
 * // This file is part of the "Hyrax Data Server" project.
 * //
 * //
 * // Copyright (c) 2018 OPeNDAP, Inc.
 * // Author: Nathan David Potter  <ndp@opendap.org>
 * //
 * // This library is free software; you can redistribute it and/or
 * // modify it under the terms of the GNU Lesser General Public
 * // License as published by the Free Software Foundation; either
 * // version 2.1 of the License, or (at your option) any later version.
 * //
 * // This library is distributed in the hope that it will be useful,
 * // but WITHOUT ANY WARRANTY; without even the implied warranty of
 * // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * // Lesser General Public License for more details.
 * //
 * // You should have received a copy of the GNU Lesser General Public
 * // License along with this library; if not, write to the Free Software
 * // Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
 * //
 * // You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
 * /////////////////////////////////////////////////////////////////////////////
 */

package opendap.auth;

import org.jdom.Element;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.ReentrantLock;
import java.util.regex.Pattern;

/**
 * Created by ndp on 10/1/14.
 */
public class MembershipRulesManager {

    private static ConcurrentHashMap<String, AuthGroup> authGroups;
    private static ConcurrentHashMap<String, HashSet<String>> authRoles;
    private static ReentrantLock configLock;
    static {
        authGroups = new ConcurrentHashMap<>();
        authRoles = new ConcurrentHashMap<>();
        configLock = new ReentrantLock();
    }

    public static void init(Element config) throws ConfigurationException {

        configLock.lock();
        try {
            for (Object o : config.getChildren("group")) {
                Element groupElem = (Element) o;
                addGroup(groupElem);
            }

            for (Object o : config.getChildren("role")) {
                Element roleElem = (Element) o;
                addRole(roleElem);
            }
        }
        finally {
            configLock.unlock();
        }
    }

    private static void addGroup(Element groupElem) throws ConfigurationException {
        String gid = groupElem.getAttributeValue("id");
        if (gid == null) {
            throw new ConfigurationException("init() - Every <group> MUST have an \"id\" attribute.");
        }

        authGroups.putIfAbsent(gid, new AuthGroup(gid));
        AuthGroup authGroup = authGroups.get(gid);

        Iterator userItr = groupElem.getChildren("user").iterator();
        if(!userItr.hasNext()){
            throw new ConfigurationException("init() - Every <group> MUST have at least one <user> element.");
        }

        while (userItr.hasNext()) {
            Element user = (Element) userItr.next();
            String uid = user.getAttributeValue("id");
            String uidPatternStr = user.getAttributeValue("idPattern");
            if ((uid==null && uidPatternStr == null) || (uid!=null && uidPatternStr!=null)) {
                throw new ConfigurationException("init(): Every <user> MUST have either an \"id\" attribute " +
                        "or an \"idPattern\" attribute, but NOT both.");
            }
            if(uid!=null){
                // We turn the uid string into a literal pattern to prevent it from actually being
                // interpreted as a pattern...
                uidPatternStr = Pattern.quote(uid);
            }

            String authContext = user.getAttributeValue("authContext");
            String authContextPatternStr = user.getAttributeValue("authContextPattern");
            if ((authContext==null && authContextPatternStr == null) || (authContext!=null && authContextPatternStr!=null)) {
                throw new ConfigurationException("init(): Every <user> MUST have either an \"authContext\" attribute " +
                        "or an \"authContextPattern\" attribute, but NOT both.");
            }

            if(authContext!=null){
                // We turn the authContext string into a literal pattern to prevent it from actually being
                // interpreted as a pattern...
                authContextPatternStr = Pattern.quote(authContext);
            }
            authGroup.addUserPattern(uidPatternStr,authContextPatternStr);
        }
    }


    private static void addRole(Element roleElem) throws ConfigurationException {
        String rid = roleElem.getAttributeValue("id");
        if (rid == null) {
            throw new ConfigurationException("init(): Every <role> must have an \"id\" attribute.");
        }

        Iterator uItr = roleElem.getChildren("group").iterator();
        if(uItr.hasNext()){
            authRoles.putIfAbsent(rid, new HashSet<String>() );
            HashSet<String> members = authRoles.get(rid);
            if (members == null)
                throw new ConfigurationException("addRole() - Unable to process role: '"+rid+"'");

            while (uItr.hasNext()) {
                Element user = (Element) uItr.next();
                String gid = user.getAttributeValue("id");
                if (gid == null) {
                    throw new ConfigurationException("init(): Every <group> must have an \"id\" attribute.");
                }
                members.add(gid);
            }
        }
    }



    private static Vector<String> getUserGroups(String uid, String authContext){
        Vector<String> groupMemberships = new Vector<>();
        for(AuthGroup authGroup : authGroups.values()){
            if(authGroup.isMember(uid, authContext)){
                groupMemberships.add(authGroup.name());
            }
        }
        return groupMemberships;
    }

    public static HashSet<String> getUserRoles(String uid, String authContext){

        HashSet<String> userRoles = new HashSet<>();

        Vector<String> userGroups = getUserGroups(uid,authContext);
        for(String roleId: authRoles.keySet()){
            HashSet<String> members = authRoles.get(roleId);
            for(String gid: userGroups) {
                if (members.contains(gid)) {
                    userRoles.add(roleId);
                    break;
                }
            }
        }
         return userRoles;
    }

}
